/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid library code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

#include <dirent.h>
#include <mntent.h>
#include <stdlib.h>
#include <linux/hdreg.h>
#include <sys/ioctl.h>
#include "internal.h"
#include "scsi.h"

/*
 * subdirectory below sysfs moint point holding the 
 * subdirectory hierarchies of all block devices.
 */
#define	BLOCK 		"/block"

/* Find sysfs mount point */
static char *find_sysfs_mp(void)
{
	char *ret = NULL;
	struct mntent *ent;
	FILE *mtab;

	if (!(mtab = fopen(_PATH_MOUNTED, "r")))
		LOG_ERR(0, "Failed to open /etc/mtab");

	while ((ent = getmntent(mtab))) {
		if (!strcmp(ent->mnt_fsname, "sysfs")) {
			ret = ent->mnt_dir;
			break;
		}
	}

	fclose(mtab);
	return ret;
}

/* Make up an absolute sysfs path given a relative one */
static char *mk_sysfs_path(char const *path) {
	char *ret = NULL;
	char *sysfs_mp;

	if (!(sysfs_mp = find_sysfs_mp()))
		LOG_ERR(NULL, "Failed to find sysfs mount point");

	if (!(ret = dbg_malloc(strlen(sysfs_mp) + strlen(path) + 1)))
		LOG_ERR(NULL, "Failed to allocate sysfs path");

	sprintf(ret, "%s%s", sysfs_mp, path);
	return ret;
}

/* Allocate disk_info struct and keep the device path */
struct disk_info *alloc_disk_info(char *path)
{
	struct disk_info *di;

	if ((di = dbg_malloc(sizeof(*di)))) {
		memset(di, 0, sizeof(*di));
		if ((di->path = dbg_strdup(path)))
			INIT_LIST_HEAD(&di->list);
		else {
			dbg_free(di);
			di = NULL;
		}
	}

	return di;
}

/* Free disk_info structure */
void free_disk_info(struct disk_info *di)
{
	if (di->rd)
		free_dev(di->rd);

	dbg_free(di->path);
	dbg_free(di);
}

/* Free all registered disk_infos. */
void free_disk_infos(struct lib_context *lc)
{
	struct list_head *elem, *tmp;

	list_for_each_safe(elem, tmp, &lc->disk_infos) {
		list_del(elem);
		free_disk_info(list_entry(elem, struct disk_info, list));
	}

	if (!list_empty(&lc->disk_infos))
		log_err("FATAL: disk_info list not empty");
}

/*
 * Create device node path, allocate disk_info and ioctl for disk geometry.
 */
/* REMOVEME: Test with dm devices ? */
#ifdef	DMRAID_TEST
static int dm_device(char *name)
{
	return !strncmp(name, "dm-", 3);
}

/* Fake a SCSI serial number by reading it from a file. */
static int get_dm_serial(struct disk_info **di, char *path)
{
	int i, ret = 0;
	FILE *f;
	char *serial, buffer[32];

	if (!(serial = dbg_malloc(strlen(path) + strlen(".serial") + 1)))
		LOG_ERR(0, "Failed allocating dm device serial path for %s",
			path);

	sprintf(serial, "%s.serial", path);
	if (!(f = fopen(serial, "r")))
		goto free;

	if (!fgets(buffer, 32, f))
		goto out;

	for (i = strlen(buffer) - 1; i > 0 && buffer[i] == '\n'; i--)
		buffer[i] = '\0';

	if (!((*di)->serial = dbg_malloc(strlen(buffer) + 1)))
		goto out;

	strcpy((*di)->serial, buffer);
	ret = 1;
  out:
	fclose(f);
  free:
	dbg_free(serial);

	return ret;
}
#endif

static int alloc_di_ioctl(struct disk_info **di, char *name)
{
	int fd, ret = 0, sector_size;
	char *path;

	if (!(path = dbg_malloc(strlen(name) + strlen(_PATH_DEV) + 1)))
		LOG_ERR(0, "Failed allocating block device path");

	sprintf(path, "%s%s", _PATH_DEV, name);
	if ((fd = open(path, O_RDONLY)) == -1)
		goto free;

	if (!(*di = alloc_disk_info(path)))
		goto close;

	if (ioctl(fd, BLKSSZGET, &sector_size))
		sector_size = DMRAID_SECTOR_SIZE;

	if (sector_size != DMRAID_SECTOR_SIZE) {
		log_err("Unsupported sector size %d on %s.", sector_size, path);
		goto free1;
	}

	/* REMOVEME: Test with dm devices ? */
#ifdef	DMRAID_TEST
	if (dm_device(name)) {
		(*di)->geometry.heads	  = DM_HEADS;
		(*di)->geometry.sectors	  = DM_SECTORS;
		(*di)->geometry.cylinders = 0; /* get_size() adjusts this */
		(*di)->geometry.start	  = 0;
		if (!get_dm_serial(di, path))
			log_err("Missing dm serial file for %s", path);
	} else {
#endif
#ifdef	HDIO_GETGEO_BIG /* Linux 2.4 */
		if (ioctl(fd, HDIO_GETGEO_BIG, &(*di)->geometry)) {
#endif
			struct hd_geometry geometry;
	
			/* Fall back if no big geometry */
			if (ioctl(fd, HDIO_GETGEO, &geometry)) {
				log_err("Failed to retrieve disk geometry "
					"from %s", path);
				goto free1;
			}
	
			(*di)->geometry.heads	  = geometry.heads;
			(*di)->geometry.sectors	  = geometry.sectors;
			(*di)->geometry.cylinders = geometry.cylinders;
			(*di)->geometry.start	  = geometry.start;
			get_scsi_serial(di, fd);
#ifdef	HDIO_GETGEO_BIG
		}
#endif
#ifdef	DMRAID_TEST
	}
#endif

	(*di)->sectors = (*di)->geometry.cylinders * (*di)->geometry.heads * (*di)->geometry.sectors;
	ret = 1;
	goto close;

   free1:
	free_disk_info(*di);
	*di = NULL;
   close:
	close(fd);
    free:
	dbg_free(path);

	return ret;
}

/* Read the size in sectors from the sysfs "size" file */
static int _get_size(struct disk_info **di, const char *path, char *name)
{
	int ret = 0;
	char *size_path;
	FILE *f;

	if (!(size_path = dbg_malloc(strlen(path) + strlen(name) + 7)))
		LOG_ERR(0, "Failed allocating sysfs block device path");

	sprintf(size_path, "%s/%s/size", path, name);
	if (!(f = fopen(size_path, "r"))) {
		log_err("Failed opening %s", size_path);
		goto free;
	}

	if (fscanf(f, "%" PRIu64, &(*di)->sectors) != 1)
		log_err("Failed reading disk size for %s from sysfs", path);
	else
		ret = 1;

	fclose(f);

   free:
	dbg_free(size_path);
	return ret;
}

/* Support function to adjust crappy disk geometries */
static int adjust_geometry(struct disk_info *di)
{
	unsigned int heads = di->geometry.heads;
	unsigned int sectors = di->geometry.sectors;
	unsigned long long ci;

	while (--heads) {
		ci = di->sectors / heads / sectors;
		if ((double) di->sectors / heads / sectors == (double) ci) {
			di->geometry.heads = heads;
			di->geometry.cylinders = ci;
			return 1;
		}
	}

	return 0;
}

/* Are we interested in this device ? */
static int _interested(char *name)
{
	/*
	 * Whole IDE and SCSI disks only.
	 * Include dm devices for testing.
	 */
	return ((!isdigit(name[strlen(name)-1]) &&
	        *(name+1) == 'd' &&
	        (*name == 'h' || *name == 's'))
#ifdef	DMRAID_TEST
		|| dm_device(name)
#endif
	       ) ? 1 : 0;
}

/* Get device geometry (heads, sectors, cylinders) and size */
static int _need_adjustment(struct disk_info *di)
{
	return (di->sectors / di->geometry.heads / di->geometry.sectors) *
	    	di->geometry.heads * di->geometry.sectors != di->sectors;
}

static int get_size(struct lib_context *lc, const char *path,
		    char *name, int sysfs)
{
	struct disk_info *di = NULL;

	if (!_interested(name))
		return 1;

	if (!alloc_di_ioctl(&di, name) ||
	    (sysfs && !_get_size(&di, path, name))) {
		if (di) {
			free_disk_info(di);
			return 1;
		}

		return 0;
	}

	/*
	 * FIXME: workaround for crappy geometry numbers
	 *	  from ioctl() in alloc_di_ioctl() :-(
	 *
	 * Can happen either from bogus kernel or
	 * from intentionally using dm devices for test.
	 */
	if (_need_adjustment(di)) {
		if (adjust_geometry(di))
			log_verbose(2, "Adjusted geometry for %s (%" PRIzu
				    "/%" PRIzu "%" PRIzu ")",
				    name, di->geometry.cylinders,
				    di->geometry.heads, di->geometry.sectors);
		else
			log_verbose(2, "Couldn't adjust geometry for %s",
				    name);
	}

	list_add(&di->list, &lc->disk_infos);
	return 1;
}

/*
 * Find disk devices in sysfs or directly
 * in /dev on Linux 2.4) and keep information.
 */
int discover_devices(struct lib_context *lc)
{
	int sysfs = 1, ret = 0;
	const char *path;
	DIR *d;
	struct dirent *de;

	if (!(path = mk_sysfs_path(BLOCK))) {
		sysfs = 0;
		path = _PATH_DEV;
		log_print("Carrying on with %s", path);
	}

	if (!(d = opendir(path))) {
		log_err("Failed opening path %s", path);
		goto out;
	}

	while ((de = readdir(d)))
		get_size(lc, path, de->d_name, sysfs);

	closedir(d);
	ret = 1;

  out:
	return ret;
}
