/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid library code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

#include "internal.h"

/* Display information about a block device */
static void log_disk(struct lib_context *lc, struct disk_info *di)
{
	log_print(lc->opts.column ? "%s,%u,%u,%u,%" PRIu64 ",\"%s\"" :
				    "%s: %u/%u/%u, %12" PRIu64 " total, \"%s\"",
		  di->path, di->geometry.heads,
		  di->geometry.sectors, di->geometry.cylinders,
		  di->sectors, di->serial ? di->serial : "N/A");
}

/* Display information about a RAID device */
static void log_rd(struct lib_context *lc, struct dmraid_dev *rd)
{
	if (!rd)
		return;

	if (rd->sectors) {
		const char *type =
			rd->fmt->type ? get_type(rd->fmt->type(lc, rd)) :
					"unsupported";

		if (!type)
			type = "unknown";

		log_print(lc->opts.column ? "%s,%s,%s,%s,%" PRIu64 ",%" PRIu64 :
					    "%s: %-18s, \"%s\", %7s, %" PRIu64
					    " sectors, data@ %" PRIu64,
			  rd->di->path, rd->name, rd->fmt->name, type,
		 	  rd->sectors, rd->offset);
	} else
		log_print(lc->opts.column ? "%s,%s,free" :
					    "%s: \"%s\" format, not used",
			  rd->di->path, rd->fmt->name);
}

/* Log native RAID device information. */
static void log_rd_native(struct lib_context *lc, struct dmraid_dev *rd)
{
	if (!rd)
		return;

	if (rd->fmt->log) {
		rd->fmt->log(lc, rd);
		log_print(" ");
	} else
		log_print("\"%s\" doesn't support native logging of RAID "
			  "device information", rd->fmt->name);
}

/* Dispatch log functions. */
static void log_device(struct lib_context *lc, struct disk_info *di,
		       enum dev_type type)
{
	if (type == RAID)
		log_rd(lc, di->rd);
	else if (type == NATIVE)
		log_rd_native(lc, di->rd);
	else
		log_disk(lc, di);
}

/* Display information about a dmraid format handler */
static void log_format(struct dmraid_format *fmt)
{
	log_print("%-6s : %s", fmt->name, fmt->descr);
}

/* Display information about found block devices */
static void process_dis(void (*f)(struct lib_context *lc, struct disk_info *di,
				  enum dev_type arg),
				  struct lib_context *lc, enum dev_type type)
{
	struct disk_info *di;

	list_for_each_entry(di, &lc->disk_infos, list)
		f(lc, di, type);
}

void display_devices(struct lib_context *lc, enum dev_type type)
{
	int devs;

	if ((devs = count_devices(lc, type))) {
		log_verbose(1, "%s device%s discovered:\n",
			    (type == RAID || type == NATIVE) ? "RAID" : "Block",
			    devs == 1 ? "" : "s");
		process_dis(log_device, lc, type);
	}
}

static void _display_set(struct lib_context *lc, struct dmraid_set *rs)
{
	uint64_t sectors;
	struct dmraid_dev *rd;
	const char *type;
	const char *status;

	if (rs->type == t_group && !lc->opts.group)
		return;

	if (!(type = get_set_type(lc, rs)))
		type = "unknown";

	if (!(status = get_status(rs)))
		status = "unknown";

	sectors = calc_total_sectors(rs);
	if (lc->opts.column)
		log_print("%s,%" PRIu64 ",%u,%s,%s,%u",
			  rs->name, sectors, rs->stride,
			  type, status, count_devs(rs));
	else {
		log_print("name  : %s", rs->name);
		log_print("size  : %" PRIu64, sectors);
		if (rs->stride)
			log_print("stride: %u", rs->stride);
		log_print("type  : %s", type);
		log_print("status: %s", status);
		log_print("dev   : %d", count_devs(rs));

	}

	if (lc->opts.column && lc->opts.column < 2)
		return;

	list_for_each_entry(rd, &rs->devs, list)
		log_rd(lc, rd);
}

/* FIXME: Spock, do something better (neater). */
void display_set(struct lib_context *lc, void *v,
		 enum active_type active, int top)
{
	struct dmraid_set *rs = v;
	struct dmraid_set *r;
	int dmstatus = dm_status(rs);

	if (rs->type != t_group) {
		if ((active == A && !dmstatus) ||
		    (active == I && dmstatus))
		return;
	}

	if (!lc->opts.column) {
		if (rs->type == t_group && !lc->opts.group)
			log_print ("\n*** Group superset %s", rs->name);
		else {
			log_print("%s %s%s%set",
		  		top ? "--->" : "\n***",
		  		rs->status == s_inconsistent ?
					      "*Inconsistent* " : "",
		  		dm_status(rs) ? "Active " : "",
		  		!list_empty(&rs->sets) ?
				"Supers" : (top ? "Subs" : "S"));
		}
	}

	_display_set(lc, rs);

	list_for_each_entry(r, &rs->sets, list)
		display_set(lc, r, active, ++top);
}

/*
 * Display information about supported RAID metadata formats
 * (ie. registered format handlers)
 */
static void _list_formats(struct lib_context *lc, enum fmt_type type)
{
	struct format_list *fmt_list;

	list_for_each_entry(fmt_list, &lc->formats, list) {
		if (type == fmt_list->fmt->fmt_type)
			log_format(fmt_list->fmt);
	}
}

int list_formats(struct lib_context *lc, int arg)
{
	if (!list_empty(&lc->formats))
		log_verbose(1, "Supported Metadata Formats:\n");

	_list_formats(lc, FMT_RAID);
	_list_formats(lc, FMT_PARTITION);

	log_print("(0): Discover, (+): Discover+Activate");
	return 1;
}
