/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid library code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

/*
 * Intel Software RAID metadata format handler.
 *
 * isw_read() profited from Carl-Daniel Hailfinger's raiddetect code.
 *
 * Profited from the Linux 2.4 iswraid driver by
 * Boji Tony Kannanthanam and Martins Krikis.
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "isw.h"
#include <utils.h>

static const char *handler = "isw";

/*
 * Make up RAID set name from family_num and volume name.
 */
static size_t __name(struct isw *isw, struct isw_dev *dev,
		     char *str, size_t len)
{
	return snprintf(str, len, "isw_%u%s%s", isw->family_num,
			dev ? "_" : "", dev ? (char *) dev->volume : "") + 1;
}

static char *_name(struct isw *isw, struct isw_dev *dev)
{
        size_t len;
        char *ret;

        len = __name(isw, dev, NULL, 0);
        if ((ret = dbg_malloc(len + 1))) {
                __name(isw, dev, ret, len);
		mk_alpha(ret,
			 strlen(ret) -
			 (dev ? strlen((char *) dev->volume) - 1 : 0));
        } else
                log_err("%s: Failed allocating name string", handler);

        return ret;
}

/* Find a disk table slot by serial number. */
static struct isw_disk *get_disk(struct isw *isw, struct disk_info *di)
{
	unsigned int i;

	for (i = 0; i < isw->num_disks; i++) {
		if (!strcmp(di->serial, isw->disk[i].serial))
			return &isw->disk[i];
	}

	return NULL;
}

/*
 * Retrieve status of device.
 *
 * FIXME: is this sufficient to cover all state ?
 */
static enum status __status(unsigned int status)
{
	if ((CONFIGURED_DISK & status) &&
	    (USABLE_DISK & status))
		return s_operational;

	return s_broken;
}

static enum status _status(struct dmraid_dev *rd)
{
	struct isw *isw = rd->meta_areas[0].area;
	struct isw_disk *disk = get_disk(isw, rd->di);

	if (!disk)
		LOG_ERR(s_undef, "%s: Error finding disk table slot for %s",
			handler, rd->di->path);

	return __status(disk->status);
}

/*
 * Mapping of Intel types to generic types.
 */
static struct types types[] = {
        { ISW_T_RAID_0, t_raid0},
        { ISW_T_RAID_1, t_raid1},
        { 0, t_undef},
};

/* Neutralize disk type. */
static enum type isw_type(struct lib_context *lc, struct dmraid_dev *rd)
{
	struct isw_dev *dev = rd->private;

	if (!dev)
		return t_group;

	return rd_type(types,
		       (unsigned int) dev->vol.map.raid_level);
}

/*
 * Generate checksum of Raid metadata for mpb_size/sizeof(u32) words
 * (checksum field itself ignored for this calculation).
 */
static uint32_t _checksum(struct isw *isw)
{
	uint32_t end = isw->mpb_size / sizeof(end), ret = 0;
	uint32_t *p = (uint32_t*) isw;

	while (end--)
		ret += *p++;

	return ret - isw->check_sum;
}

/* Advance to the next isw_dev from a given one. */
static struct isw_dev *advance_raiddev(struct isw_dev *dev)
{
	int i; /* map size correction */
	struct isw_vol *vol = &dev->vol;
	struct isw_map *map = &vol->map;

	/* Correction */
	i = (map->num_members - 1) * sizeof(map->disk_ord_tbl);
	dev++; /* off by i bytes for sure */
	dev = (struct isw_dev *) ((uint8_t *) dev + i);

	if (vol->migr_state) { /* need to add space for another map */
		map = (struct isw_map *) dev; /* Yes, it sits here! */
		i = (map->num_members - 1) * sizeof(map->disk_ord_tbl);
		dev = (struct isw_dev *) ((uint8_t *) dev + sizeof(*map) + i);
	}

	return dev;
}

/* Return isw_dev by table index. */
static struct isw_dev *raiddev(struct isw *isw, unsigned int i)
{
	struct isw_dev *dev = (struct isw_dev*) &isw->disk[isw->num_disks];

	while(i--)
		dev = advance_raiddev(dev);

	return dev;
}

/*
 * Read an Intel RAID device
 */
/* Endianess conversion. */
static void to_cpu(struct isw *isw, int full)
{
	unsigned int i, j;
	struct isw_disk *disk;
	struct isw_dev *ird;

	isw->check_sum = le32_to_cpu(isw->check_sum);
	isw->mpb_size = le32_to_cpu(isw->mpb_size);
	isw->family_num = le32_to_cpu(isw->family_num);
	isw->generation_num = le32_to_cpu(isw->generation_num);

	if (!full)
		return;

	for (i = 0; i < isw->num_disks; i++) {
		disk = &isw->disk[i];

		disk->totalBlocks = le32_to_cpu(disk->totalBlocks);
		disk->scsiId = le32_to_cpu(disk->scsiId);
		disk->status = le32_to_cpu(disk->status);
	}

	for (i = 0; i < isw->num_raid_devs; i++) {
		ird = raiddev(isw, i);

		/* RAID device. */
		ird->SizeLow = le32_to_cpu(ird->SizeLow);
		ird->SizeHigh = le32_to_cpu(ird->SizeHigh);
		ird->status = le32_to_cpu(ird->status);
		ird->reserved_blocks = le32_to_cpu(ird->reserved_blocks);

		/* RAID volume only has 8 bit members. */
	
		/* RAID map. */
		ird->vol.map.pba_of_lba0 = 
			le32_to_cpu(ird->vol.map.pba_of_lba0);
		ird->vol.map.blocks_per_member = 
			le32_to_cpu(ird->vol.map.blocks_per_member);
		ird->vol.map.num_data_stripes = 
			le32_to_cpu(ird->vol.map.num_data_stripes);
		ird->vol.map.blocks_per_strip = 
			le16_to_cpu(ird->vol.map.blocks_per_strip);
		for (j = 0; j < ird->vol.map.num_members; j++) {
			ird->vol.map.disk_ord_tbl[j] = 
				le16_to_cpu(ird->vol.map.disk_ord_tbl[j]);
		}
	}
}

static int setup_rd(struct dmraid_dev *rd, struct isw *isw,
		    struct disk_info *di, loff_t isw_sboffset);
static struct dmraid_dev *isw_read(struct lib_context *lc, struct disk_info *di)
{
	size_t size = ISW_DISK_BLOCK_SIZE;
	loff_t mpbblocks, isw_sboffset = ISW_CONFIGOFFSET;
	struct isw *isw, *isw_big;
	struct dmraid_dev *rd = NULL;

	if (!(isw = alloc_private(handler, size)))
		goto out;

	if (!read_dev(di->path, (void *) isw, isw_sboffset, size)) {
		log_err("%s: failed reading first sector of ISW superblock "
			"for %s", handler, di->path);
		goto bad;
	}

	/* Check Signature */
	to_cpu(isw, 0);
	if (strncmp(isw->sig, MPB_SIGNATURE, sizeof(MPB_SIGNATURE) - 1))
		goto bad;

	/* Check version info, older versions supported */
	if (strncmp(isw->sig + sizeof(MPB_SIGNATURE) - 1,
		   MPB_VERSION_RAID2, sizeof(MPB_VERSION_RAID2) - 1) > 0) {
		log_print("%s: superblock found, but version is too high",
			  handler);
		goto bad;
	}

	/* Get the rounded up value for mpbblocks */
	mpbblocks = ((isw->mpb_size + ISW_DISK_BLOCK_SIZE - 1) /
		    ISW_DISK_BLOCK_SIZE);
	size = mpbblocks * ISW_DISK_BLOCK_SIZE;

	/* Allocate memory for extend Intel superblock */
	if (!(isw_big = alloc_private(handler, size)))
		goto bad;

	/* Save the first metadata sector */
	memcpy(isw_big, isw, ISW_DISK_BLOCK_SIZE);
	dbg_free(isw);
	isw = isw_big;

	/* Calculate offset to extended Intel superblock and try to read it */
	isw_sboffset -= (mpbblocks - 1) * ISW_DISK_BLOCK_SIZE;
	if (!read_dev(di->path, (void *) isw + ISW_DISK_BLOCK_SIZE,
		      isw_sboffset, size - ISW_DISK_BLOCK_SIZE)) {
		log_err("%s: failed reading extended superblock for %s",
			handler, di->path);
		goto bad;
	}

	/* Calculate superblock checksum */
	to_cpu(isw, 1);
	if (isw->check_sum != _checksum(isw)) {
		log_err("%s: extended superblock for %s has wrong checksum",
			handler, di->path);
			goto bad;
	}

	file_metadata(lc, handler, di, isw, size);

	if (!(rd = alloc_dev()))
		goto bad;

	if (setup_rd(rd, isw, di, isw_sboffset))
		goto out;

	free_dev(rd);
	rd = NULL;

   bad:
	dbg_free(isw);

   out:
	return rd;
}

/*
 * Write an Intel Software RAID device.
 */
static int isw_write(struct lib_context *lc, struct dmraid_dev *rd, int erase)
{
	int ret;
	struct isw *isw = rd->meta_areas[0].area;

	to_disk(isw, 1);
	ret = write_metadata(handler, rd, 0, erase);
	to_cpu(isw, 1);
	return ret;
}

#ifdef DMRAID_NATIVE_LOG
/*
 * Log native information about an ISW RAID device.
 */
#define	P(f, x, y...) \
	do { \
		log_print("0x%03x " f, \
			  (unsigned int) &x - (unsigned int) isw, y); \
	} while(0)

#define	DP(f, x) do { P(f, x, x);} while(0)

static void isw_log(struct lib_context *lc, struct dmraid_dev *rd)
{
	unsigned int d, i;
	struct isw *isw = (struct isw*) rd->meta_areas[0].area;
	struct isw_dev *ird;

	log_print("%s (%s):", rd->di->path, handler);
	P("sig: \"%*s\"", isw->sig, MAX_SIGNATURE_LENGTH, isw->sig);
	DP("check_sum: %u", isw->check_sum);
	DP("mpb_size: %u", isw->mpb_size);
	DP("family_num: %u", isw->family_num);
	DP("generation_num: %u", isw->generation_num);
	DP("reserved[0]: %u", isw->reserved[0]);
	DP("reserved[1]: %u", isw->reserved[1]);
	DP("num_disks: %u", isw->num_disks);
	DP("num_raid_devs: %u", isw->num_raid_devs);
	DP("fill[0]: %u", isw->fill[0]);
	DP("fill[1]: %u", isw->fill[1]);
	for (i = 0; i < ISW_FILLERS; i++) {
		if (isw->filler[i])
        		P("filler[%i]: %u",
				  isw->filler[i], i, isw->filler[i]);
	}

	/* Disk table. */
	for (d = 0; d < isw->num_disks; d++) {
		if (!isw->disk[d].totalBlocks)
			continue;

		P("disk[%u].serial: \"%*s\"",
			  isw->disk[d].serial, d,
			  MAX_RAID_SERIAL_LEN, isw->disk[d].serial);
		P("disk[%u].totalBlocks: %u",
			  isw->disk[d].totalBlocks,
			  d, isw->disk[d].totalBlocks);
		P("disk[%u].scsiId: 0x%x",
			  isw->disk[d].scsiId, d, isw->disk[d].scsiId);
		P("disk[%u].status: 0x%x",
			  isw->disk[d].status, d, isw->disk[d].status);
		for (i = 0; i < ISW_DISK_FILLERS; i++) {
			if (isw->disk[d].filler[i])
                		P("disk[%u].filler[%u]: %u",
				  isw->disk[d].filler[i], d, i,
				  isw->disk[d].filler[i]);
		}
	}

	/* RAID device/volume table. */
	for (d = 0; d < isw->num_raid_devs; d++) {
		ird = raiddev(isw, d);

		/* RAID device */
		P("isw_dev[%u].volume: \"%*s\"",
		  ird->volume, d, MAX_RAID_SERIAL_LEN, ird->volume);
		P("isw_dev[%u].SizeHigh: %u",
		  ird->SizeHigh, d, ird->SizeHigh);
		P("isw_dev[%u].SizeLow: %u",
		  ird->SizeLow, d, ird->SizeLow);
		P("isw_dev[%u].status: 0x%x",
		  ird->status, d, ird->status);
		P("isw_dev[%u].reserved_blocks: %u",
		  ird->reserved_blocks, d, ird->reserved_blocks);
		for (i = 0; i < ISW_DEV_FILLERS; i++) {
			if (ird->filler[i])
                		P("isw_dev[%u].filler[%u]: %u",
				  ird->filler[i], d, i,
				  ird->filler[i]);
		}

		/* RAID volume */
		for (i = 0; i < 2; i++) {
			if (ird->vol.reserved[i])
                		P("isw_dev[%u].vol.reserved[%u]: %u",
				  ird->vol.reserved[i], d, i,
				  ird->vol.reserved[i]);
		}
		P("isw_dev[%u].vol.migr_state: %u",
		  ird->vol.migr_state, d, ird->vol.migr_state);
		P("isw_dev[%u].vol.migr_type: %u",
		  ird->vol.migr_type, d, ird->vol.migr_type);
		P("isw_dev[%u].vol.dirty: %u",
		  ird->vol.dirty, d, ird->vol.dirty);
		P("isw_dev[%u].vol.fill[0]: %u",
		  ird->vol.fill[0], d, ird->vol.fill[0]);
		for (i = 0; i < 5; i++) {
			if (ird->vol.filler[i])
                		P("isw_dev[%u].vol.filler[%u]: %u",
				  ird->vol.filler[i], d, i,
				  ird->vol.filler[i]);
		}

		/* RAID map */
		P("isw_dev[%u].vol.map.pba_of_lba0: %u",
		  ird->vol.map.pba_of_lba0, d,
		  ird->vol.map.pba_of_lba0);
		P("isw_dev[%u].vol.map.blocks_per_member: %u",
		  ird->vol.map.blocks_per_member, d,
		  ird->vol.map.blocks_per_member);
		P("isw_dev[%u].vol.map.num_data_stripes: %u",
		  ird->vol.map.num_data_stripes, d,
		  ird->vol.map.num_data_stripes);
		P("isw_dev[%u].vol.map.blocks_per_strip: %u",
		  ird->vol.map.blocks_per_strip, d,
		  ird->vol.map.blocks_per_strip);
		P("isw_dev[%u].vol.map.map_state: %u",
		  ird->vol.map.map_state, d,
		  ird->vol.map.map_state);
		P("isw_dev[%u].vol.map.raid_level: %u",
		  ird->vol.map.raid_level, d,
		  ird->vol.map.raid_level);
		P("isw_dev[%u].vol.map.num_members: %u",
		  ird->vol.map.num_members, d,
		  ird->vol.map.num_members);
		for (i = 0; i < 3; i++) {
			if (ird->vol.map.reserved[i])
                		P("isw_dev[%u].vol.map.reserved[%u]: %u",
				  ird->vol.map.reserved[i], d, i,
				  ird->vol.map.reserved[i]);
		}
		for (i = 0; i < 7; i++) {
			if (ird->vol.map.filler[i])
                		P("isw_dev[%u].vol.map.filler[%u]: %u",
				  ird->vol.map.filler[i], d, i,
				  ird->vol.map.filler[i]);
		}
		for (i = 0; i < isw->num_disks; i++) {
			P("isw_dev[%u].vol.map.disk_ord_tbl[%u]: 0x%x",
			  ird->vol.map.disk_ord_tbl[i], d, i,
			  ird->vol.map.disk_ord_tbl[i]);
		}
	}
}
#undef	P
#undef	DP
#endif

/*
 * Group an Intel SW RAID disk into potentially
 * multiple RAID sets and RAID disks.
 *
 * rs contains the top-level group RAID set on entry and shall
 * be returned on success (or NULL on error).
 */
/* Create a RAID device to map a volumes segment. */
static struct dmraid_dev *_create_rd(struct dmraid_dev *rd,
				     struct isw *isw, struct isw_dev *dev)
{
	struct dmraid_dev *r;

	/* FIXME: FAILED_MAP etc. */
	if (dev->vol.map.map_state) {
		log_print("%s: unsupported map state 0x%x on %s for %s",
			  handler, dev->vol.map.map_state, rd->di->path,
			  (char *) dev->volume);
		return NULL;
	}

	if (!(r = alloc_dev()))
		return NULL;

	r->di = rd->di;
        r->name = _name(isw, dev);
	r->fmt = rd->fmt;
	
	r->offset = dev->vol.map.pba_of_lba0;
	r->sectors = dev->vol.map.blocks_per_member;

	r->private = (void*) dev;
	r->status = s_operational;

	return r;
}

/* Find an Intel RAID set or create it. */
static struct dmraid_set *_find_create_rs(struct lib_context *lc,
					  struct list_head *list,
					  struct dmraid_dev *rd,
					  struct isw_dev *dev)
{
	struct dmraid_set *rs;

	if (!(rs = find_set(lc, rd->name, 1))) {
		if (!(rs = alloc_set()))
			return NULL;
	
		rs->name   = rd->name;
		rs->status = s_setup;
		if (dev) {
			rs->type  = isw_type(lc, rd);
			rs->stride = dev->vol.map.blocks_per_strip;
		} else
			rs->type = t_group;

		list_add_tail(&rs->list, list);
	}

	return rs;
}

static struct dmraid_set *group_rd(struct lib_context *lc,
				   struct dmraid_set *rs_group,
				   struct dmraid_dev *rd_meta)
{
	unsigned int d;
	struct isw *isw = rd_meta->meta_areas[0].area;
	struct isw_dev *dev;
	struct dmraid_dev *rd;
	struct dmraid_set *rs;

	/* Loop the device/volume table. */
	for (d = 0; d < isw->num_raid_devs; d++) {
		dev = raiddev(isw, d);

		/* FIXME: sufficient ? */
		if (dev->status) {
			log_err("%s: %s volume is broken",
				handler, dev->volume);
			continue;
		}

		if (!(rd = _create_rd(rd_meta, isw, dev)))
			goto out;

		if (!(rs = _find_create_rs(lc, &rs_group->sets, rd, dev))) {
			free_dev(rd);
			goto out;
		}

		/* FIXME: sorted!!! */
		list_add_tail(&rd->list, &rs->devs);
	}

	return rs_group;

  out:
	return NULL;
}

/* Add an Intel SW RAID device to a set */
static struct dmraid_set *isw_group(struct lib_context *lc,
				    struct dmraid_dev *rd)
{
	struct dmraid_set *rs;

	/*
	 * Once we get here, an Intel SW RAID disk containing a metadata area
	 * with a volume table has been discovered by isw_read.
	 */
	/* Check if a top level group RAID set already exists. */
	if (!(rs = _find_create_rs(lc, &lc->raid_sets, rd, NULL)))
		return NULL;

	/* Add the whole underlying RAID device to the group set. */
	list_add_tail(&rd->list, &rs->devs);

	/*
	 * We need to run through the volume table and create a RAID set and
	 * RAID devices hanging off it for every volume,
	 * so that the activate code is happy.
	 *
	 * A pointer to the top-level group RAID set
	 * gets returned or NULL on error.
	 */
	return group_rd(lc, rs, rd);
}

/* Check an Intel Software RAID set. */
static int _isw_check(struct dmraid_set *rs)
{
	unsigned int devs = 0;
	struct dmraid_dev *rd;
	struct isw_dev *dev;

	list_for_each_entry(rd, &rs->devs, list) {
		dev = rd->private;
		/* FIXME: more status checks ? */
		if (!dev->status)
			devs++;
	}

	if (devs)
		rs->status = s_operational;

	return devs ? 1 : 0;
}

static int isw_check(struct lib_context *lc, struct dmraid_set *rs_group)
{
	struct dmraid_set *rs;

	if (rs_group->type != t_group)
		return 0;

	list_for_each_entry(rs, &rs_group->sets, list) {
		if (!_isw_check(rs))
			return 0;
	}

	rs_group->status = s_operational;

	return 1;
}

static struct dmraid_format isw_format = {
	.name	= "isw",
	.descr	= "(+) Intel Software RAID",
	.fmt_type = FMT_RAID,
	.read	= isw_read,
	.write	= isw_write,
	.group	= isw_group,
	.check	= isw_check,
#ifdef DMRAID_NATIVE_LOG
	.log	= isw_log,
#endif
	.type	= isw_type,
};

/* Register this format handler with the format core */
int register_isw(struct lib_context *lc)
{
	return register_format(lc, &isw_format);
}

/*
 * Set the RAID device contents up derived from the Intel ones.
 *
 * This is the first one we get with here and we potentially need
 * to create many in case of multiple Intel SW RAID devices on the RAID disk.
 */
static int setup_rd(struct dmraid_dev *rd, struct isw *isw,
		    struct disk_info *di, loff_t isw_sboffset)
{
	rd->di = di;
        rd->name = _name(isw, NULL);
	rd->fmt = &isw_format;

	rd->offset = ISW_DATAOFFSET;
	rd->sectors = isw_sboffset / ISW_DISK_BLOCK_SIZE; /* FIXME: Correct ? */

	if (!(rd->meta_areas = alloc_meta_areas(rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = rd->sectors;
	rd->meta_areas[0].size = di->sectors - rd->sectors;
	rd->meta_areas[0].area = (void*) isw;

	rd->status = _status(rd);

	return 1;
}
