/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

#include "internal.h"
#include "ondisk.h"

/*
 * Register a RAID device format handler.
 */
int register_format(struct lib_context *lc, struct dmraid_format *fmt)
{
	struct format_list *fl;

	if (!(fl = dbg_malloc(sizeof(*fl))))
		return 0;

	fl->fmt = fmt;
	list_add_tail(&fl->list, &lc->formats);

	return 1;
}

/* Return format name */
const char *get_format_name(void *fmt)
{
	return ((struct dmraid_format *) fmt)->name;
}

/*
 * Register all format handlers.
 *
 * I use an array because of the growing number...
 */
static struct register_fh {
	int (*func)(struct lib_context *lc);
} register_fh[] = {
#include "register.h"
	{ NULL},
};

void unregister_formats(struct lib_context *lc)
{
	struct list_head *elem, *tmp;
	struct format_list *fl;

	list_for_each_safe(elem, tmp, &lc->formats) {
		list_del(elem);
		dbg_free(list_entry(elem, typeof(*fl), list));
	}

	INIT_LIST_HEAD(&lc->formats);
}

int register_formats(struct lib_context *lc)
{
	int ret = 1;
	struct register_fh *fh;

	for (fh = register_fh; fh->func; fh++) {
		if (!(ret = fh->func(lc)))
			goto bad;
	}

	return 1;

  bad:
	/* Clean up in case of error */
	log_err("Failed registering format");
	unregister_formats(lc);

	return ret;
}

/* Allocate private space in format handlers (eg, for on-disk metadata). */
void *alloc_private(const char *name, size_t size)
{
	void *ret;

	if ((ret = dbg_malloc(size)))
		memset(ret, 0, size);
	else
		log_err("Failed allocating %s metadata", name);

	return ret;
}

/* Allocate metadata sector array in format handlers. */
void *alloc_meta_areas(struct dmraid_dev *rd, const char *name, unsigned int n)
{
	void *ret;
	size_t size = n * sizeof(*rd->meta_areas);

	if ((ret = dbg_malloc(size))) {
		memset(ret, 0, size);
		rd->areas = n;
	} else
		log_err("Failed allocating %s metadata areas array", name);

	return ret;
}

/* Simple metadata write function for format handlers */
static int _write_metadata(const char *handler, struct dmraid_dev *rd,
			   int idx, int erase)
{
	int ret = 0;
	void *p, *tmp;

	if (idx >= rd->areas)
		goto out;

	p = tmp = rd->meta_areas[idx].area;
	if (erase &&
	    !(p = alloc_private(handler, rd->meta_areas[idx].size)))
		goto out;

	ret = write_dev(rd->di->path, (void *) p,
			rd->meta_areas[idx].offset << 9,
			rd->meta_areas[idx].size);

	if (p != tmp)
		dbg_free(p);

  out:
	return ret;
}

int write_metadata(const char *handler, struct dmraid_dev *rd,
		      int idx, int erase)
{
	unsigned int i;

	if (idx > -1)
		return _write_metadata(handler, rd, idx, erase);

	for (i = 0; i < rd->areas; i++) {
		if (!_write_metadata(handler, rd, i, erase))
			return 0;
	}

	return 1;
}

/* Check list of RAID sets. */
int check_rs(struct lib_context *lc, struct dmraid_set *rs,
	     int (*check)(struct lib_context *lc, struct dmraid_set *r))
{
	unsigned int broken, inconsistent, sets = 0;
	struct dmraid_set *r;

	rs->status = s_operational;
	if (list_empty(&rs->sets))
		return 1;

	broken = inconsistent = 0;
	list_for_each_entry(r, &rs->sets, list) {
		sets++;
		if (!check(lc, r))
			return 0;

		if (r->status == s_broken)
			broken++;

		if (r->status == s_inconsistent)
			inconsistent++;
	}

	if (broken || inconsistent)
		rs->status = s_inconsistent;

	if (broken == sets)
		rs->status = s_broken;

	return 1;
}

/* Check if format identifier is valid. */
int check_valid_format(struct lib_context *lc, char *name)
{
	struct format_list *fl;

	list_for_each_entry(fl, &lc->formats, list) {
		if (!strcmp(name, get_format_name(fl->fmt)))
			return 1;
	}

	return 0;
}


/* Update RAID set state based on operational devices. */
void set_status(struct dmraid_set *rs, unsigned int devs,
		unsigned int operational)
{
	if (operational)
		rs->status = (operational < devs) ? s_inconsistent :
						    s_operational;
	else
		rs->status = s_broken;
}
