/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

#include "internal.h"

/* Linked list of RAID sets. */
static LIST_HEAD_INIT(raid_sets);

/*
 * Levels definitions.
 *
 * Must be in the same order as in the type enum !!!
 */
const char *ascii_types[] = {
	NULL,
	"GROUP",
	"partition",
	"spare",
	"linear",
	"stripe",
	"mirror",
	/* RAID types below not supported (yet) */
	"raid3",
	"raid4",
	"raid5",
};

/*
 * Stacked types definitions.
 *
 * Order dependent on get_stack_type() below!
 */
const char *stacked_ascii_types[] = {
	"raid10",
	"raid30",
	"raid40",
	"raid50",
	"raid01",
	"raid03",
	"raid04",
	"raid05",
};

/*
 * Status definitions.
 *
 * Must be in the same order as in the status enum !!!
 */
const char *status[] = {
	NULL,
	"setup",
	"operational",
	"broken",
	"nosync",
	"inconsistent",
};

/* Fetch the respective ASCII string off the types array. */
const char *get_type(enum type l)
{
	return l >= ARRAY_SIZE(ascii_types) ? NULL : ascii_types[l];
}

/* Check, if a RAID set is stacked. */
static int stacked_set(struct dmraid_set *rs)
{
	return list_empty(&rs->sets) ? 0 : 1;
}

/* Return the RAID type of a stacked RAID set (eg, raid10). */
static const char *get_stacked_type(void *v)
{
	unsigned int i;
	struct dmraid_set *rs = v;
	enum type type = rs->type;

	if (type == t_raid0) {
		if (list_empty(&rs->sets))
			return NULL;

		i =  ARRAY_SIZE(stacked_ascii_types) / 2 - 1;
		type = list_entry(rs->sets.next, struct dmraid_set, list)->type;
	} else
		i = 0;

	switch (type) {
	case t_raid5:
		i++;
	case t_raid4:
		i++;
	case t_raid3:
		i++;
	case t_raid1:
		break;

	default:
		return NULL;
	}

	return stacked_ascii_types[i];
}

/* Return the ASCII type for a RAID set. */
const char *get_set_type(struct lib_context *lc, void *v)
{
	struct dmraid_set *rs = v;

	if (rs->type != t_group && stacked_set(rs))
		return get_stacked_type(rs);

	return get_type(rs->type);
}

/* Fetch the respective ASCII string off the types array. */
const char *get_status(void *rs)
{
	enum status s = ((struct dmraid_set *) rs)->status;

	return s >= ARRAY_SIZE(status) ? NULL : status[s];
}

/*
 * Calculate the size of the set by recursively
 * summing the size of the devices in the subsets.
 *
 * Pay attention to RAID > 0 types.
 */
uint64_t calc_total_sectors(struct dmraid_set *rs)
{
        uint64_t sectors = 0;
        struct dmraid_set *r;
        struct dmraid_dev *rd;

	if (rs->type != t_group) {
		list_for_each_entry(r, &rs->sets, list) {
	                sectors += calc_total_sectors(r);
			if (rs->type == t_raid1)
				break;
		}
	}

	list_for_each_entry(rd, &rs->devs, list) {
                sectors += rd->sectors;
		if (rs->type == t_raid1)
			break;
	}

	if (rs->type == t_raid3 || rs->type == t_raid5) {
		sectors -= list_empty(&rs->sets) ?
			   list_entry(rs->devs.next, typeof(*rd), list)->sectors :
			   calc_total_sectors(list_entry(rs->sets.next, typeof(*rs), list));
	} 

        return sectors;
}

/* Count devices in a set recursively. */
unsigned int count_devs(struct dmraid_set *rs)
{
	unsigned int ret = 0;
	struct dmraid_set *r;
	struct dmraid_dev *rd;

	list_for_each_entry(r, &rs->sets, list) {
		if (rs->type != t_group)
			ret += count_devs(r);
	}

	list_for_each_entry(rd, &rs->devs, list)
		ret++;

	return ret;
}

/* Allocate/Free RAID device (member of a RAID set). */
struct dmraid_dev *alloc_dev(void)
{
	struct dmraid_dev *ret;

	if ((ret = dbg_malloc(sizeof(*ret)))) {
		memset(ret, 0, sizeof(*ret));
		INIT_LIST_HEAD(&ret->list);
	} else
		log_err("Failed allocating RAID device");

	return ret;
}

void free_dev(struct dmraid_dev *rd)
{
	if (rd->name)
		dbg_free(rd->name);

	if (rd->meta_areas)
		dbg_free(rd->meta_areas);

	dbg_free(rd);
}

/* Allocate/Free RAID set. */
struct dmraid_set *alloc_set(void)
{
	struct dmraid_set *ret;

	if ((ret = dbg_malloc(sizeof(*ret)))) {
		memset(ret, 0, sizeof(*ret)); /* Clear needed! */
		INIT_LIST_HEAD(&ret->list);
		INIT_LIST_HEAD(&ret->sets);
		INIT_LIST_HEAD(&ret->devs);
	} else
		log_err("Failed allocating RAID set");

	return ret;
}

void free_set(struct dmraid_set *rs)
{
	dbg_free(rs);
}

/* Return != 0 in case of a partitioned RAID set. */
int partitioned_set(struct lib_context *lc, void *rs)
{
	return ((struct dmraid_set*) rs)->type == t_partition;
}

/* Return != 0 in case of a partitioned RAID set. */
int base_partitioned_set(struct lib_context *lc, void *rs)
{
	return ((struct dmraid_set*) rs)->has_partitions;
}

/* Remove a set recursively. */
/* FIXME: remove partition code in favour of kpartx ? */
void remove_set(struct lib_context *lc, struct dmraid_set *rs)
{
	struct list_head *elem, *tmp;
	struct dmraid_dev *rd;

	list_for_each_safe(elem, tmp, &rs->sets)
		remove_set(lc, list_entry(elem, struct dmraid_set, list));

	list_for_each_safe(elem, tmp, &rs->devs) {
		list_del(elem);
		rd = list_entry(elem, struct dmraid_dev, list);

		/* Prevent free_disk_infos() from freeing again. */
		rd->di->rd = NULL;

		if (partitioned_set(lc, rs))
			free_disk_info(rd->di);

		free_dev(rd);
	}

	list_del(&rs->list);
	free_set(rs);
}

/* Free all sets registered */
void free_sets(struct lib_context *lc)
{
	struct list_head *elem, *tmp;

	list_for_each_safe(elem, tmp, &lc->raid_sets)
		remove_set(lc, list_entry(elem, struct dmraid_set, list));

	if (!list_empty(&lc->raid_sets))
		log_err("FATAL: lib context RAID set list not empty");
}

/* Sort a device into a device list. */
void list_add_dev_sorted(struct dmraid_set *rs, struct dmraid_dev *new,
			 int (*f)(struct dmraid_dev *rd,
				  struct dmraid_dev *new))
{
	struct dmraid_dev *rd;

	list_for_each_entry(rd, &rs->devs, list) {
		/*
		 * Add in at the beginning or where the metadata
		 * format handler function tells us to.
		 */
		if (!f || f(rd, new)) {
			list_add(&new->list, &rd->list);
			return;
		}
	}

	list_add_tail(&new->list, &rs->devs);
}

/* Return RAID set name. */
const char *get_set_name(struct lib_context *lc, void *rs)
{
	return ((struct dmraid_set *) rs)->name;
}

/* Find RAID set by name (recursive if decend != 0). */
static struct dmraid_set *_find_set(struct lib_context *lc,
				    struct list_head *list,
				    const char *name, int decend) {
	struct dmraid_set *rs;

	if (list_empty(list) || !name)
		return NULL;

	log_dbg(2, "_find_set: searching %s", name);

	list_for_each_entry(rs, list, list) {
		if (!strcmp(rs->name, name))
			goto out;
	}

	if (decend) {
		list_for_each_entry(rs, list, list) {
			struct dmraid_set *r;

			if ((r = _find_set(lc, &rs->sets, name, decend))) {
				rs = r;
				goto out;
			}
		}
	}

	return NULL;

  out:
	log_dbg(1, "_find_set: found %s", name);
	return rs;
}

struct dmraid_set *find_set(struct lib_context *lc,
			    const char *name, int decend)
{
	return _find_set(lc, &lc->raid_sets, name, decend);
}

/* Return # of raid sets build */
static unsigned int _count_sets(struct lib_context *lc)
{
	int ret = 0;
	struct list_head *elem;

	list_for_each(elem, &lc->raid_sets)
		ret++;

	return ret;
}

/*
 * Count devices found
 */
static unsigned int _count_devices(struct lib_context *lc, enum dev_type type)
{
	unsigned int ret = 0;
	struct disk_info *di;

	list_for_each_entry(di, &lc->disk_infos, list) {
		if (DEVICE == type)
			ret++;
		else if (((RAID|NATIVE) & type) && di->rd)
			ret++;
	}

	return ret;
}

unsigned int count_devices(struct lib_context *lc, enum dev_type type)
{
	return type == SET ? _count_sets(lc) : _count_devices(lc, type);
}

/*
 * Support function for format handlers:
 *
 * Return neutralized RAID type for given mapping array (linear, raid0, ...)
 */
enum type rd_type(struct types *t, unsigned int type)
{
	for (; t->unified_type != t_undef; t++) {
		if (type == t->type)
			return t->unified_type;
	}

	return t_undef;
}

/*
 * Read RAID metadata off a device by trying
 * all registered format handlers in turn.
 */
static struct dmraid_dev *_dmraid_read(struct lib_context *lc,
				       struct disk_info *di,
				       struct dmraid_format *fmt)
{
	struct dmraid_dev *rd;

	log_verbose(2, "%s: %-6s discovering", di->path, fmt->name);

	if ((rd = fmt->read(lc, di))) {
		log_verbose(2, "%s: %s metadata discovered",
			    di->path, fmt->name);
		rd->fmt = fmt;
	}

	return rd;
}

static int _interested(struct dmraid_format *fmt, const char *format,
		       int partitions)
{
	if (fmt->fmt_type == FMT_PARTITION) {
		if (!partitions)
			return 0;
	} else if (partitions)
		return 0;

	else if (format && strcmp(format, fmt->name))
		return 0;

	return 1;
}

static struct dmraid_dev *dmraid_read(struct lib_context *lc,
				      struct disk_info *di, char const *format,
				      int partitions)
{
	struct format_list *fl;
	struct dmraid_dev *rd = NULL, *rd_tmp = NULL;

	list_for_each_entry(fl, &lc->formats, list) {
		if (_interested(fl->fmt, format, partitions) &&
		    (rd_tmp = _dmraid_read(lc, di, fl->fmt))) {
			if (rd) {
				log_err("%s: WARNING, \"%s\" and \"%s\" "
					"formats discovered "
					"(using %s)!",
					di->path, rd_tmp->fmt->name,
					rd->fmt->name, rd->fmt->name);
				free_dev(rd_tmp);
			} else
				rd = rd_tmp;
		}
	}

	if (rd_tmp && rd_tmp != rd) {
		free_dev(rd);
		rd = NULL;
	}

	return rd;
}

/*
 * Write RAID metadata to a device.
 */
static int dmraid_write(struct lib_context *lc,
			struct dmraid_dev *rd, int erase)
{
	int ret = 0;
	struct dmraid_format *fmt = rd->fmt;

	if (fmt->write) {
		log_verbose(1, "%s %s",
			    erase ? "Erasing metadata on" :
				    "Writing metadata to %s",
			    rd->di->path);
		ret = fmt->write(lc, rd, erase);
	} else
		log_err("Format \"%s\" doesn't support writing metadata",
			fmt->name);

	return ret;
}

/*
 * Group RAID device into a RAID set.
 */
static struct dmraid_set *dmraid_group(struct lib_context *lc,
				       struct dmraid_dev *rd)
{
	return rd->fmt->group(lc, rd);
}

/* Check that device names are members of devices list. */
static int _device_hit(struct disk_info *di, char **devices)
{
	char **dev;

	if (!devices || !*devices)
		return 1;

	for (dev = devices; *dev; dev++) {
		if (!strcmp(*dev, di->path))
			return 1;
	}

	return 0;
}

/*
 * Discover RAID devices.
 */
void discover_raid_devices(struct lib_context *lc, char **fmt_names, char **devices)
{
	unsigned int i;
	struct disk_info *di;
	struct dmraid_dev *rd;

	list_for_each_entry(di, &lc->disk_infos, list) {
		if (_device_hit(di, devices)) {
			i = 0;
			do {
				if ((rd = dmraid_read(lc, di, fmt_names ? fmt_names[i] : NULL, 0)))
					di->rd = rd;
			} while (fmt_names && fmt_names[++i]);
		}
	}
}

/*
 * Discover partitions on RAID sets.
 *
 * FIXME: remove partition code in favour of kpartx ?
 */
void discover_partitions(struct lib_context *lc)
{
	char *path;
	struct disk_info *di;
	struct dmraid_dev *rd;
	struct dmraid_set *rs, *r;

	list_for_each_entry(rs, &lc->raid_sets, list) {
		if (rs->has_partitions ||
		    partitioned_set(lc, rs) ||
		    !dm_status(rs))
			continue;

		if (!(path = mkdm_path(rs->name)))
			return;

		if (!(di = alloc_disk_info(path)))
			goto bad;

		dbg_free(path);
		di->sectors = calc_total_sectors(rs);
		if (!(rd = dmraid_read(lc, di, NULL, 1))) {
			free_disk_info(di);
			continue;
		}

		/*
		 * WARNING: partition group function returns
		 * a dummy pointer because of the creation of multiple
		 * RAID sets (one per partiton) it does.
		 * We don't want to access that 'pointer'!
		 */
		if ((r = dmraid_group(lc, rd))) {
			log_verbose(1, "Created partitioned RAID set(s) for %s",
				    rd->di->path);
			rs->has_partitions = 1;
		} else {
			log_err("Failed adding %s to RAID set", path);
			free_disk_info(di);
		}
	}

	return;

  bad:
	dbg_free(path);
}

/* Get format handler of RAID set */
static struct dmraid_format *_get_fmt(struct dmraid_set *rs)
{
	struct dmraid_dev *rd;

	while (!list_empty(&rs->sets))
		rs = list_entry(rs->sets.next, typeof(*rs), list);

	rd = list_entry(rs->devs.next, typeof(*rd), list);
	return rd ? rd->fmt : NULL;
}

/*
 * Check metadata consistency of raid sets.
 */
static void check_sets(struct lib_context *lc)
{
	struct list_head *elem, *tmp;
	struct dmraid_set *rs;
	struct dmraid_format *fmt;

	list_for_each_safe(elem, tmp, &lc->raid_sets) {
		rs = list_entry(elem, typeof(*rs), list);
		if (!(fmt = _get_fmt(rs)))
			continue;

		if (!fmt->check) {
			log_err("Fatal: Format \"%s\" lags a check method!",
				fmt->name);
			remove_set(lc, rs);
			continue;
		}

		if (!fmt->check(lc, rs)) {
			log_err("RAID set \"%s\" is inconsistent", rs->name);
			/*
			 * FIXME: check needed if degraded activation
			 *        is sensible.
			 */
			if (rs->type == t_raid1)
				log_err("Check of mirrored set \"%s\" failed, "
					"keeping it", rs->name);
			else {
				log_err("Check of \"%s\" failed, removing set",
					rs->name);
				remove_set(lc, rs);
			}
		}
	}

	return;
}

/*
 * Build RAID set(s)
 *
 *	name = NULL  : build all sets
 *	name = String: build just the one set
 *
 *	format = NULL  : try discovering RAID devices using
 *                       all registered formats
 *	format = String: try discovering RAID devices using the
 *                       specified format only
 *
 */
int build_set(struct lib_context *lc, char *name)
{
	struct disk_info *di;
	struct dmraid_dev *rd;
	struct dmraid_set *rs;

	if (find_set(lc, name, 0))
		LOG_ERR(0, "RAID set %s already exists\n", name);

	list_for_each_entry(di, &lc->disk_infos, list) {
		if (!(rd = di->rd) ||
		    (name && strncmp(rd->name, name, strlen(name))))
			continue;
	
		if ((rs = dmraid_group(lc, rd)))
			log_verbose(2, "Added %s to RAID set %s",
				    rd->di->path, rs->name);
		else {
			if (rd->sectors)
				log_err("Failed adding %s to RAID set",
					rd->di->path);
			else
				log_print("Skipping spare %s", rd->di->path);

			/* Need to find the failed set and remove it. */
			if ((rs = find_set(lc, rd->name, 0))) {
				log_err("Removing set %s", rs->name);
				remove_set(lc, rs);
			}
		}
	}

	/* Check sanity of grouped RAID sets. */
	check_sets(lc);

	return 1;
}

/* Process function on RAID set(s) */
static void process_set(int (*func)(struct lib_context *l, void *v, int a),
			struct lib_context *lc, void *rs, int arg)
{
	if (!partitioned_set(lc, rs))
		func(lc, rs, arg);
}

/* FIXME: remove partition code in favour of kpartx ? */
static void
process_partitioned_set(int (*func)(struct lib_context *l, void *v, int a),
			struct lib_context *lc, void *rs, int arg)
{
	if (partitioned_set(lc, rs) && !base_partitioned_set(lc, rs))
		func(lc, rs, arg);
}

void process_sets(int (*func)(struct lib_context *l, void *v, int a),
		  struct lib_context *lc, int arg, int partitions)
{
	struct dmraid_set *rs;
	void (*p)(int (*f)(struct lib_context *l, void *v, int a),
		  struct lib_context *l, void *v, int a) =
		partitions ?  process_partitioned_set : process_set;

	list_for_each_entry(rs, &lc->raid_sets, list)
		p(func, lc, rs, arg);
}

/*
 * Write RAID set metadata to devices.
 */
int write_set(struct lib_context *lc, void *v)
{
	int ret = 0;
	struct dmraid_set *r, *rs = v;
	struct dmraid_dev *rd;

	/* Decend hierarchy */
	list_for_each_entry(r, &rs->sets, list) {
		/*
		 * FIXME: does it make sense to try the rest of the subset
		 *	  in case we fail writing one ?
		 */
		if (!write_set(lc, (void *) r))
			log_err("Failed writing subset %s, continuing",
				r->name);
	}

	/* Write metadata to devices */
	list_for_each_entry(rd, &rs->devs, list) {
		if (!(ret = dmraid_write(lc, rd, 0)))
			break;
	}

	return ret;
}

/*
 * Support function for format handlers:
 *
 * File RAID metadata for analysis.
 */
static size_t _name(char *str, size_t len, char *path, const char *handler)
{
	char *s;
	
	s = strrchr(path, '/');
	s = s ? s + 1 : path;
	return snprintf(str, len, "%s_%s.dat", s, handler) + 1;
}

void file_metadata(struct lib_context *lc, const char *handler,
		   struct disk_info *di, void *data, size_t size)
{
	if (lc->opts.dump) {
		int fd;
		size_t len;
		char *name = NULL;
	
		len = _name(NULL, 0, di->path, handler) + 1;
		if (!(name = dbg_malloc(len))) {
			log_err("Error allocation metadata file memory");
			return;
		}

		_name(name, len, di->path, handler);
		if ((fd = open(name, O_CREAT|O_TRUNC|O_WRONLY, 0600)) == -1) {
			log_err("Error opening metadata file %s", name);
			goto free;
		}

		log_verbose(1, "WRITING %s", name);
		if (write(fd, data, size) != size) {
			log_err("Error writing metadata file %s", name);
			goto free;
		}
	
		close(fd);
		return;

  free:
		dbg_free(name);
		close(fd);
	}
}

/*
 * Erase ondisk metadata.
 */
int erase_metadata(struct lib_context *lc)
{
	struct disk_info *di;

	list_for_each_entry(di, &lc->disk_infos, list) {
		if (di->rd &&
		    yes_no_prompt("Do you really want to erase \"%s\" ondisk "
				  "metadata on %s",
				  di->rd->fmt->name, di->path) &&
		    !dmraid_write(lc, di->rd, 1))
			log_err("Error erasing ondisk metadata on %s",
				di->path);
	}

	return 1;
}
