/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this dmraid code; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

#include <getopt.h>
#include <string.h>
#include <dmraid.h>
// #include "misc/lib_context.h"
#include "commands.h"
#include "toollib.h"
#include "version.h"

/* Action flags */
enum action action = UNDEF;

/* Metadata formats identified by names to use during discovery. */
char **format_names = NULL;

/* Command line options */
static char const *short_opts = "a:p"
#ifndef	DMRAID_MINI
				"bcdDEf:ghl"
#ifdef	DMRAID_NATIVE_LOG
				"n"
#endif
				"rs::tvV"
#endif
;

static struct option long_opts[] = {
	{"activate", required_argument, NULL, 'a'},
	{"format", required_argument, NULL, 'f'},
	{"no_partitions", no_argument, NULL, 'p'},
#ifndef	DMRAID_MINI
	{"block_devices", no_argument, NULL, 'b'},
	{"display_columns", no_argument, NULL, 'c'},
	{"debug", no_argument, NULL, 'd'},
	{"dump_metadata", no_argument, NULL, 'D'},
	{"erase_metadata", no_argument, NULL, 'E'},
	{"display_group", no_argument, NULL, 'g'},
	{"help", no_argument, NULL, 'h'},
	{"list_formats", no_argument, NULL, 'l'},
#  ifdef	DMRAID_NATIVE_LOG
	{"native_log", no_argument, NULL, 'n'},
#  endif
	{"raid_devices", no_argument, NULL, 'r'},
	{"sets", optional_argument, NULL, 's'},
	{"test", no_argument, NULL, 't'},
	{"verbose", no_argument, NULL, 'v'},
	{"version", no_argument, NULL, 'V'},
#endif
	{NULL, no_argument, NULL, 0}
};

static int inc_column(struct lib_context *lc)
{
	return ++lc->opts.column;
}

static int inc_debug(struct lib_context *lc)
{
	return ++lc->opts.debug;
}

static int inc_dump(struct lib_context *lc)
{
	return ++lc->opts.dump;
}

static int inc_group(struct lib_context *lc)
{
	return ++lc->opts.group;
}

static int inc_test(struct lib_context *lc)
{
	return ++lc->opts.test;
}

static int inc_verbose(struct lib_context *lc)
{
	return ++lc->opts.verbose;
}


/* Check activate/deactivate option arguments. */
static int check_activate(struct lib_context *lc)
{
	size_t len;

	if (!optarg)
		return 0;

	str_tolower(optarg);
	len = strlen(optarg);
	if (!strncmp(optarg, "yes", len))
		action |= ACTIVATE;
	else if (!strncmp(optarg, "no", len))
		action |= DEACTIVATE;
	else
		LOG_ERR(0, "Invalid option argument for -a");

	return 1;
}

/* Check format identifier */
static int check_format(struct lib_context *lc)
{
	char *invalid;

	if (!valid_format(lc, optarg, &format_names, &invalid)) {
		log_err("Invalid format \"%s\" for -f at (see -l)", invalid);
		return 0;
	}

	return 1;
}

#ifndef	DMRAID_MINI
/* Check active/inactive option arguments. */
static int check_active(struct lib_context *lc)
{
	size_t len;

	if (!optarg)
		return 1;

	str_tolower(optarg);
	len = strlen(optarg);
	if (!strncmp(optarg, "active", len))
		action |= ACTIVE;
	else if (!strncmp(optarg, "inactive", len))
		action |= INACTIVE;
	else
		LOG_ERR(0, "Invalid option argument for -s");

	return 1;
}

/*
 * Action flag definitions for set_action()
 *
 * 'Early' options can be handled directly in set_action() by calling
 * the functions registered here (set on_set member).
 */
static struct actions actions [] = {
	{'a', UNDEF, UNDEF, 1, 1, check_activate,
	 ACTIVATE|DEACTIVATE|FORMAT|NO_PARTITIONS
#ifndef DMRAID_MINI
	|DBG|TEST|VERBOSE
#endif
	},
	{'f', FORMAT, ACTIVATE|DEACTIVATE|NATIVE_LOG|RAID_DEVICES|RAID_SETS,
	 1, 1, check_format, ACTIVATE|DEACTIVATE
#ifndef DMRAID_MINI
	|ACTIVE|INACTIVE|COLUMN|DBG|DUMP|ERASE
#  ifdef DMRAID_NATIVE_LOG
	|NATIVE_LOG
#  endif
	|RAID_DEVICES|RAID_SETS|TEST|VERBOSE
#endif
	},
	{'p', NO_PARTITIONS, ACTIVATE|DEACTIVATE, 1, 0, NULL, ACTIVATE|DEACTIVATE|FORMAT
#ifndef DMRAID_MINI
	|DBG|NO_PARTITIONS|TEST|VERBOSE
#endif
	},
#ifndef	DMRAID_MINI
	{'b', BLOCK_DEVICES, UNDEF, 0, 0, NULL, COLUMN|DBG|VERBOSE},
	{'c', COLUMN, BLOCK_DEVICES|RAID_DEVICES|RAID_SETS, 1, 1, inc_column,
	 ACTIVE|INACTIVE|BLOCK_DEVICES|DBG|FORMAT|GROUP|RAID_DEVICES|RAID_SETS},
	{'d', DBG, ACTIVATE|DEACTIVATE|BLOCK_DEVICES|LIST_FORMATS|
	  NATIVE_LOG|RAID_DEVICES|RAID_SETS, 1, 1, inc_debug,
	  ACTIVATE|DEACTIVATE|BLOCK_DEVICES|COLUMN|DUMP|ERASE|FORMAT|
	  LIST_FORMATS|RAID_SETS|NO_PARTITIONS|RAID_DEVICES|TEST|VERBOSE},
	{'D', DUMP, RAID_DEVICES, 1, 1, inc_dump,
	 DBG|FORMAT|RAID_DEVICES|VERBOSE},
	{'E', ERASE, RAID_DEVICES, 1, 0, NULL, DBG|FORMAT|RAID_DEVICES|VERBOSE},
	{'g', GROUP, RAID_SETS, 1, 1, inc_group,
	 ACTIVE|INACTIVE|DBG|COLUMN|FORMAT|RAID_SETS|VERBOSE},
	{'h', HELP, UNDEF, 0, 0, NULL, 0},
	{'l', LIST_FORMATS, UNDEF, 0, 0, NULL, DBG|VERBOSE},
#  ifdef DMRAID_NATIVE_LOG
	{'n', NATIVE_LOG, UNDEF, 1, 0, NULL, DBG|FORMAT|VERBOSE},
#  endif
	{'r', RAID_DEVICES, UNDEF, 1, 0, NULL,
	 COLUMN|DBG|DUMP|ERASE|FORMAT|VERBOSE},
	{'s', RAID_SETS, UNDEF, 1, 1, check_active,
	 ACTIVE|INACTIVE|COLUMN|DBG|FORMAT|GROUP|VERBOSE},
	{'t', TEST, ACTIVATE|DEACTIVATE, 1, 1, inc_test,
	 ACTIVATE|DEACTIVATE|DBG|FORMAT|NO_PARTITIONS|VERBOSE},
	{'v', VERBOSE, ACTIVATE|DEACTIVATE|BLOCK_DEVICES|LIST_FORMATS|
	 NATIVE_LOG|RAID_DEVICES|RAID_SETS, 1, 1, inc_verbose,
	 ACTIVATE|DEACTIVATE|COLUMN|DBG|DUMP|ERASE|BLOCK_DEVICES|FORMAT|
	 LIST_FORMATS|RAID_SETS|NO_PARTITIONS|RAID_DEVICES|TEST},
	{'V', VERSION, UNDEF, 0, 0, NULL, 0},
#endif
	{0, UNDEF, UNDEF, 0, 0, NULL, UNDEF},
};

/*
 * Set action flag and optionally call function on set.
 */
static int set_action(struct lib_context *lc, int o)
{
	struct actions *a;

	for (a = actions; a->option; a++) {
		if (o == a->option) {
			action |= a->action;	/* Set action flag. */
			a->allowed |= a->action;/* Add to allowed flags. */
			if (a->on_set)		/* Optionally call function. */
				return a->function(lc);

			break;
		}
	}

	return 1;
}

/* Check for invalid option combinations */
static int check_actions(struct lib_context *lc, char **argv)
{
	struct actions *a;

	for (a = actions; a->option; a++) {
		if (a->action & action) {
			if (a->needed != UNDEF &&
			    !(a->needed & action))
				LOG_ERR(0,
					"invalid option combination with -%c",
					a->option);

			if (~a->allowed & action)
#ifndef DMRAID_MINI
				LOG_ERR(0, "Invalid option combination"
					   " (-h for help)");
#else
				LOG_ERR(0, "Invalid option combination");
#endif

			if (!a->args && argv[optind])
				LOG_ERR(0, "No arguments allowed with -%c\n",
					a->option);
		}
	}

	if (!action)
		LOG_ERR(0, "Options missing\n");

#ifndef DMRAID_MINI
	if ((action & (DBG|VERBOSE)) == action)
		LOG_ERR(0, "More options needed with -d/-v");
#endif

	return 1;
}

/* Parse and handle the command line arguments */
int handle_args(struct lib_context *lc, int argc, char ***argv)
{
	int o, opt_idx, ret = 0;

	if (argc < 2)
#ifndef DMRAID_MINI
		LOG_ERR(0, "No arguments/options given (-h for help)\n");
#else
		LOG_ERR(0, "No arguments/options given\n");
#endif

	/* Walk the options (and option arguments) */
	while ((o = getopt_long(argc, *argv, short_opts,
				long_opts, &opt_idx)) != -1) {
		if (o == ':' || o == '?' || !set_action(lc, o))
			return 0;
	}

	/* Force deactivation of stacked partition devices. */
	/* FIXME: remove partiton code in favour of kpartx ? */
	if (DEACTIVATE & action)
		action &= ~NO_PARTITIONS;

	ret = check_actions(lc, *argv);
	*argv += optind;

	return ret;
}

/* Display help information */
#ifndef	DMRAID_MINI
static int help(struct lib_context *lc, int arg)
{
	char *c = lc->cmd;

	log_print("%s: Device-Mapper Software RAID tool\n", c);
	log_print("* = [-d|--debug]{0,} [-v|--verbose]{0,}\n");
	log_print("%s\t{-a|--activate} {y[es]/n[o]} *\n"
		  "\t[-f|--format FORMAT]\n"
		  "\t[-p|--no_partitions]\n"
		  "\t[-t|--test]\n"
		  "\t[RAID set]{0,}\n", c);
	log_print("%s\t{-b|--block_devices} *\n"
		  "\t[-c|--display_columns]\n", c);
	log_print("%s\t{-h|--help}\n", c);
	log_print("%s\t{-l|--list_formats} *\n", c);
#  ifdef	DMRAID_NATIVE_LOG
	log_print("%s\t{-n|--native_log} *\n"
		  "\t[-f|--format FORMAT]\n"
		  "\t[device-path]{0,}\n", c);
#  endif
	log_print("%s\t{-r|--raid_devices} *\n"
		  "\t[-c|--display_columns]\n"
		  "\t[-D|--dump_metadata]\n"
		  "\t[-E|--erase_metadata]\n"
		  "\t[-f|--format FORMAT]\n"
		  "\t[device-path]{0,}\n", c);
	log_print("%s\t{-s|--sets}[a[active]/i[nactive]] *\n"
		  "\t[-c|--display_columns]{0,2}\n"
		  "\t[-f|--format FORMAT]\n"
		  "\t[-g|--display_group]\n"
		  "\t[RAID set]{0,}\n", c);
	log_print("%s\t{-V/--version}\n", c);

	return 1;
}
#endif

static int version(struct lib_context *lc, int arg)
{
	char v[80];

	dm_version(v, sizeof(v));
	log_print("%s version:\t\t%s", lc->cmd, DMRAID_VERSION);
	log_print("%s library version:\t%s", lc->cmd, libdmraid_version(lc));
	log_print("device-mapper version:\t%s", v);

	return 1;
}
#endif

/*********************************************************************
 * Perform pre/post functions for requested actions.
 */
/* Post Activate/Deactivate RAID set. */
#ifndef DMRAID_MINI
/* Pre and post display_set() functions. */
static int _display_sets_arg(int arg)
{
	return (action & ACTIVE) ? A : ((action & INACTIVE) ? I : ALL);
}

static int _display_set(struct lib_context *lc, void *rs, int type)
{
	display_set(lc, rs, type, 0);
	return 1;
}

static int _display_sets(struct lib_context *lc, int type)
{
	process_sets(_display_set, lc, type, 0);
	return 1;
}

/* Pre and post display_devices functions. */
static int _display_arg(int arg)
{
	return arg;
}

static int _display_devices(struct lib_context *lc, int type)
{
	display_devices(lc, type);
	return 1;
}

static int _erase(struct lib_context *lc, int arg)
{
	return erase_metadata(lc);
}
#endif

/* Retrieve and build metadata. */
static int get_metadata(struct lib_context *lc, struct prepost *p, char **argv)
{
	if (!(M_DEVICE & p->metadata))
		return 1;

	if (!discover_devices(lc))
		return 0;

	if (!count_devices(lc, DEVICE))
		return 0;

	if (!(M_RAID & p->metadata))
		return 1;

#ifndef	DMRAID_MINI
	/* Discover RAID disks and keep RAID metadata (eg, hpt45x) */
	discover_raid_devices(lc, format_names,
		    ((NATIVE_LOG|RAID_DEVICES) & action) ? argv : NULL);
#else
	discover_raid_devices(lc, format_names, NULL);
#endif
	if (!count_devices(lc, RAID)) {
		format_error(format_names, "disk");
		return 0;
	}

	if (M_SET & p->metadata) {
		/* Group RAID sets. */
		build_sets(lc, argv);
		if (!count_devices(lc, SET)) {
			format_error(format_names, "set");
			return 0;
		}
	}

	return 1;
}

/* Definition of pre- and post functions to perform. */
struct prepost prepost[] = {
	{ ACTIVATE|DEACTIVATE, M_DEVICE|M_RAID|M_SET, NULL, 0, activate_sets },
#ifndef DMRAID_MINI
	{ BLOCK_DEVICES, M_DEVICE, _display_arg, DEVICE, _display_devices },
	{ ERASE, M_DEVICE|M_RAID, NULL, 0, _erase },
	{ HELP, M_NONE, NULL, 0, help },
	{ LIST_FORMATS, M_NONE, NULL, 0, list_formats },
#  ifdef DMRAID_NATIVE_LOG
	{ NATIVE_LOG, M_DEVICE|M_RAID, _display_arg, NATIVE, _display_devices },
#  endif
	{ RAID_DEVICES, M_DEVICE|M_RAID, _display_arg, RAID, _display_devices },
	{ RAID_SETS, M_DEVICE|M_RAID|M_SET, _display_sets_arg, 0, _display_sets },
#endif
	{ VERSION, M_NONE, NULL, 0, version },
	{ 0, M_NONE, NULL, 0, NULL },
};

/*
 * Function abstraction which takes pre- and post-function calls
 * to prepare an argument in pre to be used by post.
 *
 * perform() is the call handler for all functions which need metadata
 * as displaying, erasing and activation/deactivation of RAID sets.
 *
 * The necessary metadata about disks, RAID devices and RAID sets
 * gets automatically generated by this function.
 */
int perform(struct lib_context *lc, char **argv)
{
	struct prepost *p;

	for (p = prepost; p->action; p++) {
		if (p->action & action) {
			if (!get_metadata(lc, p, argv))
				return 0;

			return p->post(lc, p->pre ? p->pre(p->arg) : 0);
		}
	}

	return 0;
}
